@@ -26,16 +26,36 @@
 
 #define BPC_MAXPATHLEN          (2 * MAXPATHLEN)
 
-extern char BPC_PoolDir[];
-extern char BPC_CPoolDir[];
-extern char BPC_PoolDir3[];
-extern char BPC_CPoolDir3[];
-extern char BPC_TopDir[];
 extern int BPC_HardLinkMax;
 extern int BPC_PoolV3Enabled;
 extern int BPC_TmpFileUnique;
 extern int BPC_LogLevel;
 
+/*
+ * String buffer functions
+ */
+typedef struct _bpc_strBuf {
+    char *s;
+    size_t size;
+    char buffer[2048];
+    struct _bpc_strBuf *next;   /* for free list */
+} bpc_strBuf;
+
+bpc_strBuf *bpc_strBuf_new(void);
+void bpc_strBuf_init(bpc_strBuf *buf);
+void bpc_strBuf_free(bpc_strBuf *buf);
+void bpc_strBuf_resize(bpc_strBuf *buf, size_t needed);
+char *bpc_strBuf_strcpy(bpc_strBuf *buf, unsigned int offset, char *str);
+char *bpc_strBuf_strcat(bpc_strBuf *buf, unsigned int offset, char *str);
+int bpc_strBuf_vsnprintf(bpc_strBuf *buf, unsigned int offset, char *fmt, va_list args);
+int bpc_strBuf_snprintf(bpc_strBuf *buf, unsigned int offset, char *fmt, ...);
+
+extern bpc_strBuf BPC_PoolDir;
+extern bpc_strBuf BPC_CPoolDir;
+extern bpc_strBuf BPC_PoolDir3;
+extern bpc_strBuf BPC_CPoolDir3;
+extern bpc_strBuf BPC_TopDir;
+
 /*
  * Maximum length of a digest - 16 bytes for MD5, but 4 bytes of collision counting
  */
@@ -104,7 +124,7 @@ void bpc_poolRefRequestFsck(char *targetDir, int ext);
  */
 typedef struct {
     bpc_refCount_info refCnt[2];
-    char targetDir[BPC_MAXPATHLEN];
+    bpc_strBuf *targetDir;
 } bpc_deltaCount_info;
 
 void bpc_poolRefDeltaFileInit(bpc_deltaCount_info *info, char *hostDir);
@@ -158,7 +178,7 @@ typedef struct _bpc_candidate_file {
     bpc_digest digest;
     OFF_T fileSize;
     int v3File;
-    char fileName[BPC_MAXPATHLEN];
+    bpc_strBuf *fileName;
     struct _bpc_candidate_file *next;
 } bpc_candidate_file;
 
@@ -168,7 +188,7 @@ typedef struct {
     int v3File;
     OFF_T fileSize;
     bpc_digest digest;
-    char fileName[BPC_MAXPATHLEN];
+    bpc_strBuf *fileName;
 } bpc_candidate_match;
 
 typedef struct {
@@ -200,7 +220,7 @@ typedef struct {
      */
     int fdOpen;
     bpc_fileZIO_fd fd;
-    char tmpFileName[BPC_MAXPATHLEN];
+    bpc_strBuf *tmpFileName;
     /*
      * Error count
      */
@@ -239,17 +259,22 @@ void bpc_digest_append_ext(bpc_digest *digest, uint32 ext);
 void bpc_digest_digest2str(bpc_digest *digest, char *hexStr);
 void bpc_digest_str2digest(bpc_digest *digest, char *hexStr);
 int bpc_digest_compare(bpc_digest *digest1, bpc_digest *digest2);
-void bpc_digest_md52path(char *path, int compress, bpc_digest *digest);
-void bpc_digest_md52path_v3(char *path, int compress, bpc_digest *digest);
+void bpc_digest_md52path(bpc_strBuf *path, int compress, bpc_digest *digest);
+void bpc_digest_md52path_v3(bpc_strBuf *path, int compress, bpc_digest *digest);
 void bpc_digest_buffer2MD5_v3(bpc_digest *digest, uchar *buffer, size_t bufferLen);
-void bpc_fileNameEltMangle(char *path, int pathSize, char *pathUM);
-void bpc_fileNameMangle(char *path, int pathSize, char *pathUM);
+void bpc_fileNameEltMangle(bpc_strBuf *path, char *pathUM);
+void bpc_fileNameMangle(bpc_strBuf *path, char *pathUM, unsigned int i);
 void bpc_logMsgf(char *fmt, ...);
 void bpc_logErrf(char *fmt, ...);
 void bpc_logMsgGet(char **mesg, size_t *mesgLen);
 void bpc_logMsgErrorCntGet(unsigned long *errorCnt);
 void bpc_logMsgCBSet(void (*cb)(int errFlag, char *mesg, size_t mesgLen));
 
+/*
+ * Compute file digest
+ */
+int bpc_fileDigest(char *fileName, int compress, bpc_digest *digest);
+
 /*
  * Directory operations
  */
@@ -304,6 +329,7 @@ typedef struct {
 typedef struct {
     bpc_digest digest;
     ushort compress;
+    ushort needRewrite;
     /*
      * hash table of bpc_attrib_file entries, indexed by file name
      */
@@ -320,6 +346,7 @@ size_t bpc_attrib_xattrList(bpc_attrib_file *file, char *list, size_t listLen, i
 void bpc_attrib_fileInit(bpc_attrib_file *file, char *fileName, int xattrNumEntries);
 void bpc_attrib_fileDestroy(bpc_attrib_file *file);
 bpc_attrib_file *bpc_attrib_fileGet(bpc_attrib_dir *dir, char *fileName, int allocate_if_missing);
+int bpc_attrib_fileIterate(bpc_attrib_dir *dir, bpc_attrib_file **file, uint *idx);
 void bpc_attrib_fileCopyOpt(bpc_attrib_file *fileDest, bpc_attrib_file *fileSrc, int overwriteEmptyDigest);
 void bpc_attrib_fileCopy(bpc_attrib_file *fileDest, bpc_attrib_file *fileSrc);
 int bpc_attrib_fileCompare(bpc_attrib_file *file0, bpc_attrib_file *file1);
@@ -328,12 +355,13 @@ int bpc_attrib_fileCount(bpc_attrib_dir *dir);
 char *bpc_attrib_fileType2Text(int type);
 void bpc_attrib_dirInit(bpc_attrib_dir *dir, int compressLevel);
 void bpc_attrib_dirDestroy(bpc_attrib_dir *dir);
+int bpc_attrib_dirNeedRewrite(bpc_attrib_dir *dir);
 ssize_t bpc_attrib_getEntries(bpc_attrib_dir *dir, char *entries, ssize_t entrySize);
 void bpc_attrib_dirRefCount(bpc_deltaCount_info *deltaInfo, bpc_attrib_dir *dir, int incr);
 void bpc_attrib_dirRefCountInodeMax(bpc_deltaCount_info *deltaInfo, bpc_attrib_dir *dir, int incr, unsigned int *inodeMax);
-void bpc_attrib_attribFilePath(char *path, char *dir, char *attribFileName);
+void bpc_attrib_attribFilePath(bpc_strBuf *path, char *dir, char *attribFileName);
 bpc_digest *bpc_attrib_dirDigestGet(bpc_attrib_dir *dir);
-uchar *bpc_attrib_buf2file(bpc_attrib_file *file, uchar *buf, uchar *bufEnd, int xattrNumEntries);
+uchar *bpc_attrib_buf2file(bpc_attrib_file *file, uchar *buf, uchar *bufEnd, int xattrNumEntries, int *xattrFixup);
 uchar *bpc_attrib_buf2fileFull(bpc_attrib_file *file, uchar *buf, uchar *bufEnd);
 uchar *bpc_attrib_file2buf(bpc_attrib_file *file, uchar *buf, uchar *bufEnd);
 int bpc_attrib_digestRead(bpc_attrib_dir *dir, bpc_digest *digest, char *attribPath);
@@ -396,13 +424,11 @@ typedef struct {
      */
     bpc_deltaCount_info *deltaInfo;
 
-    char shareName[BPC_MAXPATHLEN];
-    int shareNameLen;
-    char shareNameUM[BPC_MAXPATHLEN];
-    char hostName[BPC_MAXPATHLEN];
-    char hostDir[BPC_MAXPATHLEN];
-    char backupTopDir[BPC_MAXPATHLEN];
-    char currentDir[BPC_MAXPATHLEN];
+    bpc_strBuf *shareName;
+    bpc_strBuf *shareNameUM;
+    bpc_strBuf *hostName;
+    bpc_strBuf *backupTopDir;
+    bpc_strBuf *currentDir;
 } bpc_attribCache_info;
 
 typedef struct {
@@ -434,6 +460,6 @@ int bpc_attribCache_deleteInode(bpc_attribCache_info *ac, ino_t inode);
 int bpc_attribCache_getDirEntryCnt(bpc_attribCache_info *ac, char *path);
 ssize_t bpc_attribCache_getDirEntries(bpc_attribCache_info *ac, char *path, char *entries, ssize_t entrySize);
 void bpc_attribCache_flush(bpc_attribCache_info *ac, int all, char *path);
-void bpc_attribCache_getFullMangledPath(bpc_attribCache_info *ac, char *path, char *dirName, int backupNum);
+void bpc_attribCache_getFullMangledPath(bpc_attribCache_info *ac, bpc_strBuf *path, char *dirName, int backupNum);
 
 #endif
